/*
	WARNING:	In the current rev, the alert(6) widget (this code)
			is nothing more than a toy.  If your filter rules
			are at all ornate, packets will be dropped at
			moderate traffic loads.  If you're serious about
			capturing packets for analysis, use the sensor(1)
			widget instead.				-spb
*/
#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_config.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"
#include "shoki_parser.h"


char *dumpfile = NULL;
u_char *pcap_userdata = 0;
pcap_handler out = pcap_print_pt;
struct bpf_program base_fcode;
extern int nfilters;
int silent = 0, timeval_kludge = 0, linux_kludge = 0, fragre = 0, callbacks = 1;
FILE *logfile;
int snaplen = DEFAULT_SNAPLEN;

char *read_file(char *fname);

int main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL;
	bpf_u_int32 localnet, netmask;
	char errbuf[PCAP_ERRBUF_SIZE];
	struct sigaction *cleansig;
	struct sigaction *childsig;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "c:C:F:f:i:r:s:")) != EOF)
		switch(op)
		{
			case 'c':
				count = atoi(optarg);
				break;
			case 'C':
				cfile = optarg;
				break;
			case 'F':
				filterlist = optarg;
				break;
			case 'f':
				currentfile = optarg;
				break;
			case 'i':
				dev = optarg;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
		}

	base_filter = get_filter(&argv[optind]);

	cleansig = malloc(sizeof(struct sigaction));
	childsig = malloc(sizeof(struct sigaction));

	childsig->sa_handler = (void *)clean_kids;
	childsig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	cleansig->sa_handler = (void *)cleanup;
	cleansig->sa_flags = SA_RESTART & SA_NOCLDSTOP;

	(void)sigaction(SIGCHLD, childsig, NULL);
	(void)sigaction(SIGTERM, cleansig, NULL);
	(void)sigaction(SIGINT, cleansig, NULL);

	((cfile == NULL) ? (parse_shoki_config(ALERT_CONFIG)) : 
		(parse_shoki_config(cfile)));
	if((pf_outformat[0] == 0) || (pf_outdelim[0] == NULL))
		(void)default_formats(pf_outformat_d, pf_outdelim_d, pf_type_d);
	if((fname != NULL) && (base_filter == NULL))
		base_filter = read_file(fname);
	if(base_filter == NULL) error("Specify a filter, dammit.");
	if(filterlist == NULL) error("Specify a filterlist, dammit.");
	(void)parse_config(filterlist);
	if(dev == NULL) error("Specify a device, dammit.");
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "alert starting");
       	(void)closelog();
	logfile = stdout;
	fflush(NULL);
	if(dumpfile != NULL)
	{
		pd = fake_pcap(pd, 0);
		(void)read_zdumps(dumpfile, out);
		exit(0);
	}
	pd = pcap_open_live(dev, snaplen, 1, 1000, errbuf);
	if(pd == NULL) error("%s", errbuf);
	localnet = 0;
	netmask = 0;
	if(pcap_lookupnet(dev, &localnet, &netmask, errbuf) < 0)
		((localnet = 0), (netmask = 0), error("%s", errbuf));
	(void)compile_filters(pd, netmask, 0);
	if(pcap_compile(pd, &base_fcode, base_filter, 1, netmask) < 0) 
		error("%s\n", pcap_geterr(pd));
	if(pcap_setfilter(pd, &base_fcode) < 0) 
		error("%s\n", pcap_geterr(pd));
	for(;;)
		(void)capture_loop();
	pcap_close(pd);
	exit(0);
}

int
cleanup(int signo)
{
	if(pd != NULL && pcap_file(pd) == NULL)
	{
		(void)fflush(stdout);
		putc('\n', stderr);
	}
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "alert shutting down");
       	(void)closelog();
	exit(0);
}

void
capture_loop()
{
	if(pcap_loop(pd, count, out, pcap_userdata) < 0)
	error("%s", pcap_geterr(pd));
}

void
clean_kids()
{
	waitpid(-1, NULL, WNOHANG);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	struct filter *filter;
	int i;

	if(!(filter = (struct filter *)calloc(1, sizeof(struct filter))))
		error("calloc:  pcap_print_pt");
	raise(SIGCHLD);
	for(i = 1; i <= nfilters; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
		{
			if(callback_pt[i] == 0) return;
			make_filter(i, filter);
			callback_handler((u_char *)sp, filter, h->caplen, h, 1);
		}
	}
	free(filter);
}
