#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"

struct doctrine *doctrine;
extern int nfilters, rnfilters;
extern int count, default_snaplen, callbacks, silent;
void diddle_filter(struct doctrine *, const struct dconfig *);

struct dconfig *dconfig_list;

#define READ_BUF_SIZE 1024

static struct { const char *name; int option; } doc_opt[] =
{
	{ "duration",	DO_DURATION },
	{ "label",	DO_LABEL },
	{ "log_all",	DO_LOGALL },
	{ "rewrite",	DO_REWRITE },
	{ "filter",	DO_FILTER },
	{ "file",	DO_FILE },
	{ "command",	DO_COMMAND },
	{ "instances",	DO_INSTANCES },
	{ "reset",	DO_RESET },
	{ NULL,		0 }
};

static struct { const char *name; int option; } ip_bits[] =
{
	{ "IP.HL.",             IP_S_HL },
	{ "IP.V.",              IP_S_V },
	{ "IP.TOS.",            IP_S_TOS },
	{ "IP.LEN.",            IP_S_LEN },
	{ "IP.ID.",             IP_S_ID },
	{ "IP.OFF.",            IP_S_OFF },
	{ "IP.TTL.",            IP_S_TTL },
	{ "IP.P.",              IP_S_P },
	{ "IP.SUM.",            IP_S_SUM },
	{ "IP.SRC.",		IP_S_SRC },
	{ "IP.DST.",            IP_S_DST },
	{ NULL,			0 }
};

#define WHITE "\t\n\r "

void
parse_doctrine_config(const char *config)
{
	FILE *f = NULL;
	char line[1024];
	int i = 0, j = 0, lnum = 0, match = -1;
	char *foo, *bar = NULL, *baz = NULL;
	struct doctrine *dlist;
	struct dconfig *dc;

	for(dc = dconfig_list; dc->next; dc = dc->next);
	if(!(dc->next = (struct dconfig *)calloc(1, sizeof(struct dconfig))))
		error("calloc:  parse_doctrine_config");
	if(!(dc->doctrine = (struct doctrine *)calloc(1, sizeof(struct doctrine))))
		error("calloc:  parse_doctrine_config");
	dlist = dc->doctrine;
	f = fopen(config, "r");
	if(!f) error("Couldn't read doctrine from %s", config);
	while(fgets(line, sizeof(line), f))
	{
		lnum++;
		foo = line + strspn(line, WHITE);
		if(foo == NULL || !*foo || *foo == '#') continue;
		bar = strdup(foo);
		foo = strtok(foo, WHITE);
		for(i=0; foo[i]; i++) foo[i] = tolower(foo[i]);
		match = -1;
		for(j=0; doc_opt[j].name && (match == -1); j++)
			if(strcmp(foo, doc_opt[j].name) == 0)
				match = doc_opt[j].option;
		if(match == -1) error("Bad doctrine option in %s, line %d",
			config, lnum);
		baz = bar + strspn(bar, foo);
		baz = baz + strspn(baz, WHITE);
		for(i=(strlen(baz) - 1); isspace(baz[i]) && i > 0; i--)
			baz[i] = '\0';
		switch(match)
		{
			case DO_DURATION:
				dc->duration = atoi(baz);
				break;
			case DO_LABEL:
				strncpy(dc->label, baz, strlen(baz));
				break;
			case DO_LOGALL:
				dc->log_all = 1;
				break;
			case DO_REWRITE:
				if(strlen(baz) > MAX_FILE_LEN)
					error("Filename too long in %s, line %d", config, lnum);
				strncpy(dc->outfile, baz, strlen(baz));
				break;
			case DO_FILTER:
				(void)parse_doctrine_filter(bar, config, lnum,
					dlist, dc);
				break;
			case DO_FILE:
				(void)parse_doctrine_file(bar, config, lnum,
					dlist, dc);
				break;
			case DO_COMMAND:
				(void)parse_doctrine_command(bar, config,
					lnum, dlist, dc);
				break;
			case DO_INSTANCES:
				dc->max_instances = atoi(baz);
				break;
		}
		if(!(dlist->next = (struct doctrine *)calloc(1, sizeof(struct doctrine))))
			error("calloc:  parse_doctrine_config");
		dlist = dlist->next;
	}
}

#define TAB "\t"

void
parse_doctrine_filter(const char *line, const char *config, int lnum, struct doctrine *dlist, struct dconfig *dc)
{
	dlist->dtype = DOCTRINE_FILTER;
	(void)parse_doctrine_rule(line, config, lnum, dlist, dc);
}

void
parse_doctrine_file(const char *line, const char *config, int lnum, struct doctrine *dlist, struct dconfig *dc)
{
	dlist->dtype = DOCTRINE_FILE;
	(void)parse_doctrine_rule(line, config, lnum, dlist, dc);
}

void
parse_doctrine_command(const char *line, const char *config, int lnum, struct doctrine *dlist, struct dconfig *dc)
{
	dlist->dtype = DOCTRINE_COMMAND;
	(void)parse_doctrine_rule(line, config, lnum, dlist, dc);
}

int
handle_doctrine(const struct pcap_pkthdr *h, const u_char *sp)
{
	const struct timeval *tvp;
	struct dconfig *dc;
	struct doctrine *d;
	struct filter *filter;
	char *foo = NULL;
	int i, j, fd, len, r, left, p;

	(const struct timeval *)tvp = &h->ts;
	if(!(filter = (struct filter *)calloc(1, sizeof(struct filter))))
		error("calloc:  handle_doctrine");

	for(dc = dconfig_list; dc->next; dc = dc->next)
	{
		if((dc->sec + dc->duration) < tvp->tv_sec)
		{
			dc->lastmatch = 0;
			dc->matches = 0;
			for(d = dc->doctrine; d->next; d = d->next)
			{
				d->matched = 0;
			}
		}
		i = dc->lastmatch + 1;
		for(d = dc->doctrine; d->next; d = d->next)
		{
			if(d->matched) continue;
			if(d->sev > i) continue;
			switch(d->dtype)
			{
				case DOCTRINE_FILTER:
					if(d->recompile)
						diddle_filter(d, dc);
					if(bpf_filter(d->fcode->bf_insns,
						(u_char *)sp, h->len,
						h->caplen))
					{
						if((d->callback != 0) &&
							(callbacks != 0))
						{
							doctrine_to_filter(d, filter);
							j = silent;
							if(dc->log_all)
								silent = 0;
							else
								silent = 1;
							if(callback_handler((u_char *)sp, filter, h->caplen, h, dc->log_all))
							{
								if(dc->lastmatch == 0)
								{
									dc->sec = tvp->tv_sec;
									dc->usec = tvp->tv_usec;
								}
								if(dc->lastmatch < i) dc->lastmatch = i;
								dc->matches++;
								if((dc->outfile != NULL) || (dc->save_packets != 0))
								{
									if(!(d->hh = (struct pcap_pkthdr *)calloc(1, sizeof(struct pcap_pkthdr))))
										error("calloc:  handle_doctrine");
									if(!(d->sp = (u_char *)calloc(1, h->caplen)))
										error("calloc:  handle_doctrine");
									memcpy(d->hh, h, sizeof(struct pcap_pkthdr));
									memcpy(d->sp, sp, h->caplen);
								}
								d->matched++;
							}
							silent = j;
						}
					}
					break;
				case DOCTRINE_FILE:
					fd = open(d->filter, O_RDONLY);
					if(fd < 0)
					{
						(void)log_msg(h, NULL, NULL, "File open of %s for doctrine rule failed", d->filter);
						continue;
					}
					p = lseek(fd, 0, SEEK_END);
					p -= d->fcount;
					if(p < 0) p = 0;
					len = lseek(fd, p, SEEK_SET);
					if(len < 0)
					{
						(void)log_msg(h, NULL, NULL, "File seek in %s for doctrine rule failed", d->filter);
						close(fd);
						continue;
					}
					if(d->fcount > READ_BUF_SIZE)
						r = READ_BUF_SIZE;
					else
						r = d->fcount;
					left = d->fcount;
					if(!(foo = (char *)calloc(1, r)))
						error("calloc:  handle_doctrine");
					while(left > 0)
					{
						len = read(fd, foo, r);
						if(len < 0)
						{
							(void)log_msg(h, NULL, NULL, "Bad read from %s in doctrine rule", d->filter);
							close(fd);
							continue;
						}
						if(regexsearch(foo, r, d->regex, strlen(d->hex)))
						{
							if(dc->lastmatch == 0)
							{
								dc->sec = tvp->tv_sec;
								dc->usec = tvp->tv_usec;
							}
							dc->lastmatch = i;
							if(d->matched == 0)
								dc->matches++;
							d->matched++;
							left = 0;
						}
						if(left > r)
							lseek(fd, (off_t)(-(r - 1)), SEEK_CUR);
						left -= r;
					}
					free(foo);
					close(fd);
					break;
				case DOCTRINE_COMMAND:
					dc->lastmatch = i;
					if(d->matched == 0)
						dc->matches++;
					d->matched++;
					break;
			}
		}
		if(dc->matches == dc->items)
		{
			dc->matches = 0;
			dc->lastmatch = 0;
			for(d = dc->doctrine; d->next; d = d->next)
			{
				free(d->hh);
				free(d->sp);
				d->matched = 0;
			}
			(void)log_msg(h, (sp + sizeof(struct ether_header)),
				NULL, "%s", dc->label);
			if(dc->outfile != NULL)
			{
			}
		}
	}
	free(filter);
	return(1);
}

void
parse_doctrine_rule(const char *line, const char *config, int lnum, struct doctrine *dlist, struct dconfig *dc)
{
	char *foo, *baz = NULL, *hex = NULL, hexen[MAX_HEX_LEN];
	struct doctrine *d;
	int i, k, t[3], re_flags = 0;

	foo = (char *)line;
	foo = strtok(foo, TAB);
	foo = strtok(NULL, TAB);
	if((foo == NULL) || !*foo)
		error("Bad filter in %s, line %d", config, lnum);
	d = dlist;
	strncpy(d->filter, foo, MAX_FILTER_LEN);
	if(!(d->fcode = (struct bpf_program *)calloc(1, sizeof(struct bpf_program))))
		error("calloc:  parse_doctrine_rule");
	foo = strtok(NULL, TAB);
	if((foo == NULL) || !*foo)
		error("Bad filter in %s, line %d after filter.", config, lnum);
	d->fcount = atoi(foo);
	foo = strtok(NULL, TAB);
	if((foo == NULL) || !*foo || *foo == '#')
		error("Bad filter in %s, line %d after count.", config, lnum);
	k = 0;
	while((baz = rindex(foo, ':')))
	{
		*baz = '\0';
		baz++;
		t[k] = atoi(baz);
		k++;
		if(k > 4)
			error("Bad filter in %s, line %d in callback.",
			config, lnum);
	}
	k--;
	if(k > 0)
	{
		if(k == 1)
			error("Bad filter in %s, line %d in callback.",
			config, lnum);
		d->d = t[k];
		d->t = t[k - 1];
		if(k > 1) d->q = t[k - 2];
		if(k > 2) d->h = t[k - 3];
	}
	if((foo == NULL) || !*foo || *foo == '#')
		error("Bad filter in %s, line %d in callback.", config, lnum);
	d->callback = get_callback(foo);
	if(d->callback == 0)
		error("Bad callback in %s, line %d.", config, lnum);
	foo = strtok(NULL, TAB);
	if((foo == NULL) || !*foo)
		error("Bad filter in %s, line %d after callback.",
		config, lnum);
	strncpy(d->id, foo, MAX_ID_LEN);
	foo = strtok(NULL, TAB);
	if((foo == NULL) || !*foo)
		error("Bad filter in %s, line %d after ID.", config, lnum);
	strncpy(d->call_args, foo, MAX_CALLARG_LEN);
	foo = strtok(NULL, TAB);
	if((foo == NULL) || !*foo)
		error("Bad filter in %s, line %d after args.", config, lnum);
	if(!(baz = (char *)calloc(1, MAX_HEX_LEN)))
		error("calloc:  parse_doctrine_rule");
	if(!(hex = (char *)calloc(1, 4)))
		error("calloc:  parse_doctrine_rule");
	for(i=0, k=0; i <= strlen(foo); i++, k++)
	{
		if(strncasecmp(&foo[i], "0x", 2) == 0)
		{
			i += 2;
			while(isxdigit(foo[i]))
			{
				memcpy(hex, &foo[i], 2);
				hex[2] = '\0';
				baz[k] = strtol(hex, '\0', 16);
				i += 2;
				k++;
			}
			i--;
			k--;
		}
		else
			memcpy(&baz[k], &foo[i], 1);
	}
	if(strncasecmp(foo, "NULL", 4) == 0)
	{
	}
	else
	{
		strncpy(hexen, baz, MAX_HEX_LEN);
		re_flags = REG_EXTENDED | REG_NOSUB;
	}
	memcpy(d->hex, hexen, MAX_HEX_LEN);
	if(!(d->regex = (regex_t *)calloc(1, sizeof(regex_t))))
		error("calloc:  parse_doctrine_rule");
	if(regcomp(d->regex, hexen, (re_flags)) != 0)
		error("Problem compiling (%s) into regex in %s, line %d",
		hexen, config, lnum);
	foo = strtok(NULL, TAB);
	if((foo == NULL) || !*foo || (*foo == '#'))
		error("Bad filter in %s, line %d, after search.", config,
		lnum);
	strncpy(d->type, foo, MAX_TYPE_LEN);
	foo = strtok(NULL, TAB);
	if((foo != NULL) && *foo && (*foo != '#'))
		d->sev = atoi(foo);
	d->recompile = check_tokens(d, config, lnum);
	dc->save_packets += d->recompile;
	dc->items++;
	d->index = rnfilters;
	rnfilters++;
	free(baz);
	free(hex);
}

int
check_tokens(const struct doctrine *d, const char *config, int lnum)
{
	char *foo = NULL, *bar;
	int i, j;
	if(!(bar = (char *)calloc(1, MAX_FILTER_LEN)))
		error("calloc:  check_tokens");
	strncpy(bar, d->filter, MAX_FILTER_LEN);
	foo = bar;
	foo = strtok(foo, WHITE);
	/* The first token should never match.  If it does, the filter
	   expression is munged. -spb */
	while((foo = strtok(NULL, WHITE)))
		for(i=0; ip_bits[i].name; i++)
			if(strncasecmp(foo, ip_bits[i].name, strlen(ip_bits[i].name)) == 0)
			{
				/* Talk about your cryptic error messages. */
				if(d->sev <= 1)
					error("Reference in doctrine rule invalide for rules numbered less than one,\nline %d in %s", lnum, config);
				j = atoi(&foo[strlen(ip_bits[i].name)]);
				if(j > d->sev)
					error("Doctrine reference needs to be lower than current rule number,\nline %d in %s", lnum, config);
				free(bar);
				return(1);
			}
	free(bar);
	return(0);
}

void
diddle_filter(struct doctrine *d, const struct dconfig *dc)
{
	struct ip *ip;
	struct doctrine *d2;
	struct bpf_program fcode;
	struct bpf_program *f;
	char *foo = NULL, *bar, *baz, filt[MAX_FILTER_LEN], tmp[MAX_FILTER_LEN];
	int i, j;

	if(d->recompile == 0) return;
	if(!(bar = (char *)calloc(1, MAX_FILTER_LEN)))
		error("calloc:  diddle_filter");
	if(!(baz = (char *)calloc(1, MAX_FILTER_LEN)))
		error("calloc:  diddle_filter");
	strncpy(bar, d->filter, MAX_FILTER_LEN);
	foo = bar;
	foo = strtok(foo, WHITE);
	if(strlen(foo) > MAX_FILTER_LEN)
		error("diddle_filter:  diddled filter too long");
	strncpy(filt, foo, MAX_FILTER_LEN);
	/* The first token should never match.  If it does, the filter
	   expression is munged. -spb */
	while((foo = strtok(NULL, WHITE)))
	{
		j = 0;
		for(i=0; ip_bits[i].name; i++)
			if(strncasecmp(foo, ip_bits[i].name, strlen(ip_bits[i].name)) == 0)
			{
				j = atoi(&foo[strlen(ip_bits[i].name)]);
				for(d2 = dc->doctrine; d2->next; d2 = d2->next)
				{
					if(d2->sev != j) continue;
					if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
						error("calloc:  diddle_filter");
					ip = (struct ip *)(d2->sp + sizeof(struct ether_header));
					strncpy(tmp, get_bit(ip, ip_bits[i].option), MAX_FILTER_LEN);
					if((strlen(filt) + strlen(tmp) + 1) > MAX_FILTER_LEN)
						error("diddle_filter:  diddled filter too long");
					strncat(filt, " ", 1);
					strncat(filt, tmp, strlen(tmp));
				}
			}
		if(j == 0)
		{
			if((strlen(filt) + strlen(foo) + 1) > MAX_FILTER_LEN)
				error("diddle_filter:  diddled filter too long");
			strncat(filt, " ", 1);
			strncat(filt, foo, strlen(foo));
		}
	}
	if(pcap_compile(dc->pd, &fcode, filt, 1, dc->netmask) < 0)
		error("pcap error(%s)\n\tin doctrine recompile of %s",
		pcap_geterr(dc->pd), filt);
	f = &fcode;
	free(d->fcode);
	if(!(d->fcode = (struct bpf_program *)calloc(1, sizeof(*f))))
		error("calloc:  diddle_filter");
	memcpy(d->fcode, f, sizeof(*f));
	free(bar);
	free(baz);
}

char *
get_bit(struct ip *ip, int option)
{
	char foo[MAX_FILTER_LEN];
	char *bar;

	switch(option)
	{
		case IP_S_HL:
			sprintf(foo, "%d", ip->ip_hl);
			break;
		case IP_S_V:
			sprintf(foo, "%d", ip->ip_v);
			break;
		case IP_S_TOS:
			sprintf(foo, "%d", ip->ip_tos);
			break;
		case IP_S_LEN:
			sprintf(foo, "%d", ip->ip_len);
			break;
		case IP_S_ID:
			sprintf(foo, "%d", ip->ip_id);
			break;
		case IP_S_OFF:
			sprintf(foo, "%d", ip->ip_off);
			break;
		case IP_S_TTL:
			sprintf(foo, "%d", ip->ip_ttl);
			break;
		case IP_S_P:
			sprintf(foo, "%d", ip->ip_p);
			break;
		case IP_S_SUM:
			sprintf(foo, "%d", ip->ip_sum);
			break;
		case IP_S_SRC:
			sprintf(foo, "%s", ipaddr_string(&ip->ip_src));
			break;
		case IP_S_DST:
			sprintf(foo, "%s", ipaddr_string(&ip->ip_dst));
			break;
		default:
			/* This could only be made better by returning
			   it in hex. -spb */
			error("Bad bit requested");
			break;
	}
	foo[strlen(foo)] = '\0';
	bar = foo;
	return(bar);
}
