#include <pcap.h>
#include <pcap-int.h>
#include <regex.h>
#include <zlib.h>
#include "shoki.h"
#include "shoki_utils.h"
#include "shoki_doctrine.h"

extern char *currentfile, *dev, *fname, *outfile, *filterlist;
extern char *ostype, *portlist, **filter_pt;
extern struct bpf_program **fcode_pt;
extern int count, default_snaplen, start, stop, verbose;
extern int query_window, query_max, query_period, nfilters, rnfilters;
extern pcap_t *pd;
extern struct bpf_program base_fcode;
extern char *base_filter;
char email_rcpt[256];
extern int linux_kludge, timeval_kludge;

char blah[MAX_FILTER_LEN];

extern struct dconfig *dconfig_list;

#ifndef SHOKI_LOG_NAME
#define SHOKI_LOG_NAME	"shoki"
#endif

int
dumbsearch(const char *foo, int foolen, const char *bar, int barlen)
{
	int i, j, k;


	if(barlen > foolen) return(0);
	i = 0;
	j = 0;
	while(i <= (foolen - barlen))
	{
		if(foo[i] != bar[j])
		{
			i++;
		}
		else
		{
			for(j = 0,  k = i; j < barlen && foo[k] == bar[j]; j++, k++);
			if(j == barlen) return(1);
			j = 0;
			i++;
		}
	}
	return(0);
}

int
regexsearch(const char *bar, int barlen, const regex_t *foo, int foolen)
{
	char *bar2;
	int i, rerr;
	char rebuf[512];

	if(!(bar2 = (char *)calloc(1, barlen + 1)))
		error("calloc:  regexsearch");
	memcpy(bar2, bar, barlen);
	/* Fuck.  Is this really necessary?  -spb */
	for(i = 0; i <= barlen; i++)
	{
		if(bar2[i] == '\0') bar2[i] = ' ';
	}
	rerr = regexec(foo, bar2, 0, 0, 0);
	free(bar2);
	if(rerr == 0) return(1);
	if(rerr == REG_NOMATCH)
	{
		return(0);
	}
	(void)regerror(rerr, foo, rebuf, 512);
	fprintf(stderr, "Regex error:  %s\n", rebuf);
	return(0);
	/* never reached */
	error("Regex error:  %s", rebuf);
	/* really never reached */
	return(0);
}

pcap_t *
fake_pcap(pcap_t *pd, int foo)
{
	char errbuf[PCAP_ERRBUF_SIZE];
	char bar[MAX_FILTER_LEN];
	int baz = 1;
	bpf_u_int32 netmask = 0;

	if(base_filter != NULL)
	{
		memcpy(&bar, &base_filter, MAX_FILTER_LEN);
		baz = 0;
	}
	pd = pcap_open_offline(FAKEFILE, errbuf);
	if(pd == NULL)
		(void)make_fake(FAKEFILE);
	pd = pcap_open_offline(FAKEFILE, errbuf);
	if(pd == NULL) error("%s", errbuf);
	if(baz == 0)
	{
		memcpy(&base_filter, &bar, MAX_FILTER_LEN);
		if(pcap_compile(pd, &base_fcode, base_filter, 1, netmask) < 0)
			error("fake_pcap:  pcap error (%s)\n\tparsing base_filter (%s)",
			pcap_geterr(pd), base_filter);
	}
	if(rnfilters > 0) compile_filters(pd, netmask, baz);
	if(foo != 0)
	{
		pcap_close(pd);
		return(NULL);
	}
	return(pd);
}

void
compile_filters(pcap_t *pd, bpf_u_int32 netmask, int get_base)
{
	compile_fcodes(pd, netmask, get_base);
	return;
}

void
compile_fcodes(pcap_t *pd, bpf_u_int32 netmask, int get_base)
{
	int i;
	struct bpf_program fcode;
	struct bpf_program *f;
	struct doctrine *d;
	struct dconfig *dc;

	if(!(fcode_pt = (struct bpf_program **)calloc(1, sizeof(struct bpf_program) * MAX_FILTERS)))
		error("calloc:  compile_fcodes");
	for(i = 1; i <= nfilters; i++)
	{
		fcode = compile_fcode(pd, i, netmask);
		f = &fcode;
		if(!(fcode_pt[i] = (struct bpf_program *)calloc(1, sizeof(*f))))
			error("calloc:  compile_fcodes");
		memcpy(fcode_pt[i], f, sizeof(*f));
	}
	for(dc = dconfig_list; dc->next; dc = dc->next)
	{
		dc->pd = pd;
		dc->netmask = netmask;
		for(d = dc->doctrine; d->next; d = d->next)
		{
			if(d->dtype != DOCTRINE_FILTER) continue;
			if(d->recompile != 0) continue;
			if(pcap_compile(pd, &fcode, d->filter, 1, netmask)
				< 0)
				error("compile_fcodes:  pcap error (%s)\nparsing (%s)",
					pcap_geterr(pd), d->filter);
			f = &fcode;
			free(d->fcode);
			if(!(d->fcode = (struct bpf_program *)calloc(1, sizeof(*f))))
				error("calloc:  compile_fcodes");
			memcpy(d->fcode, f, sizeof(*f));
		}
	}
	if(get_base == 0) return;
	if(base_filter == NULL)
	{
		if(!(base_filter = (char *)calloc(1, strlen(filter_pt[rnfilters]) + 1)))
			error("calloc:  compile_fcodes");
		strncpy(base_filter, filter_pt[rnfilters], strlen(filter_pt[rnfilters]));
	}
	if(pcap_compile(pd, &base_fcode, base_filter, 1, netmask) < 0)
		error("compile_fcodes:  pcap error (%s)\n\tparsing base_filter (%s)",
		pcap_geterr(pd), base_filter);
}

struct bpf_program
compile_fcode(pcap_t *pd, int i, bpf_u_int32 netmask)
{
	struct bpf_program fcode;
	
	if(pcap_compile(pd, &fcode, filter_pt[i], 1, netmask) < 0)
		error("compile_fcode:  pcap error (%s)\nparsing (%s)", pcap_geterr(pd), filter_pt[i]);
	return(fcode);
}

#define WHITE " \n\r\t"
void
read_zdumps(char *fnames, pcap_handler out)
{
	char *fname;
	fname = strdup(fnames);
	fname = strtok(fname, WHITE);
	if(fname == NULL) return;
	read_zdump(fname, out);
	while((fname = strtok(NULL, WHITE)) && (fname != NULL))
		read_zdump(fname, out);
}

void
read_zdump(char *zfname, pcap_handler out)
{
	char *zfile = NULL;
	u_char *sp;
	struct pcap_pkthdr *h;
	struct linux_kludge_pkthdr *kp;
	struct pcap_file_header hdr;
	int bar, swapped = 0;

	if(timeval_kludge != 0)
	{
		(void)read_wee_zdump(zfname);
		return;
	}
	if(!(h = (struct pcap_pkthdr *)calloc(1, sizeof(struct pcap_pkthdr))))
		error("calloc:  read_zdump");
	if(!(kp = (struct linux_kludge_pkthdr *)calloc(1, sizeof(struct linux_kludge_pkthdr))))
		error("calloc:  read_zdump");
	zfile = gzopen(zfname, "rb");
	if(zfile == NULL) error("Couldn't open dump (%s)", zfname);
	if(gzread(zfile, (char *)&hdr, sizeof(struct pcap_file_header)) < 0)
		error("Bad dump format");
	if(hdr.magic != TCPDUMP_MAGIC)
	{
		if(SWAPLONG(hdr.magic) != TCPDUMP_MAGIC)
			error("Bad magic in tcpdump header");
		swapped = 1;
	}
	if(!(sp = (u_char *)calloc(1, 65535)))
		error("calloc:  read_zdump");
	while((bar = gzread(zfile, (struct pcap_pkthdr *)h, sizeof(struct pcap_pkthdr))) > 0)
	{
		if(swapped == 1)
		{
			h->caplen = SWAPLONG(h->caplen);
			h->len = SWAPLONG(h->len);
			h->ts.tv_sec = SWAPLONG(h->ts.tv_sec);
			h->ts.tv_usec = SWAPLONG(h->ts.tv_usec);
		}
		if(linux_kludge != 0)
		{
			bar = gzread(zfile, (struct linux_kludge_pkthdr *)kp, sizeof(struct linux_kludge_pkthdr));
		}
		if((h->caplen > 65535) || (h->caplen < 0))
		{
			if((timeval_kludge == 0) && (linux_kludge == 0))
				error("Bogus caplen; try using -k.");
			if(timeval_kludge == 0)
				error("Bogus caplen; try using -K.");
			gzclose(zfile);
			free(h);
			free(sp);
			free(kp);
			(void)read_wee_zdump(zfname);
			return;
		}
		if((bar = gzread(zfile, (u_char *)sp, h->caplen)) < 0)
			error("Bad packet read");
		if(check_frag(h, sp) == 1)
		{
			if(verbose > 0)
				if(bpf_filter(base_fcode.bf_insns, sp, h->len,
					h->caplen))
					out(NULL, h, sp);
		}
		else
		{
			if(bpf_filter(base_fcode.bf_insns, sp, h->len,
				h->caplen))
				out(NULL, h, sp);
		}
	}
	gzclose(zfile);
	(void)frag_reassemble(0);
	free(h);
	free(sp);
	free(kp);
}

/* The size of the pcap_pkthdr structure depends on the size of time_t
   on the local architecture.  read_wee_zdump() is a kludge to allow
   architectures with sizeof(time_t) > 4 to read dumps made by
   architectures with a 32 bit time_t.  This could be handled considerably
   more elegantly.				-spb
*/
void
read_wee_zdump(char *zfname)
{
	char *zfile = NULL;
	u_char *sp;
	struct wee_pcap_pkthdr *h;
	struct pcap_pkthdr *hh;
	struct wee_linux_kludge_pkthdr *kp;
	struct pcap_file_header hdr;
	int bar, swapped = 0;

	if(!(h = (struct wee_pcap_pkthdr *)calloc(1, sizeof(struct wee_pcap_pkthdr))))
		error("calloc:  read_wee_zdump");
	if(!(hh = (struct pcap_pkthdr *)calloc(1, sizeof(struct pcap_pkthdr))))
		error("calloc:  read_wee_zdump");
	if(!(kp = (struct wee_linux_kludge_pkthdr *)calloc(1, sizeof(struct wee_linux_kludge_pkthdr))))
		error("calloc:  read_wee_zdump");
	zfile = gzopen(zfname, "rb");
	if(zfile == NULL) error("Couldn't open dump (%s)", zfname);
	if(gzread(zfile, (char *)&hdr, sizeof(struct pcap_file_header)) < 0)
		error("Bad dump format");
	if(hdr.magic != TCPDUMP_MAGIC)
	{
		if(SWAPLONG(hdr.magic) != TCPDUMP_MAGIC)
			error("Bad magic in tcpdump header");
		swapped = 1;
	}
	if(!(sp = (u_char *)calloc(1, 65535)))
		error("calloc:  read_wee_zdump");
	while((bar = gzread(zfile, (struct wee_pcap_pkthdr *)h, sizeof(struct wee_pcap_pkthdr))) > 0)
	{
		if(swapped == 1)
		{
			h->caplen = SWAPLONG(h->caplen);
			h->len = SWAPLONG(h->len);
			h->ts.tv_sec = SWAPLONG(h->ts.tv_sec);
			h->ts.tv_usec = SWAPLONG(h->ts.tv_usec);
		}
		hh->caplen = h->caplen;
		hh->len = h->len;
		hh->ts.tv_sec = (time_t)h->ts.tv_sec;
		hh->ts.tv_usec = (time_t)h->ts.tv_usec;
		if(linux_kludge != 0)
		{
			bar = gzread(zfile, (struct wee_linux_kludge_pkthdr *)kp, sizeof(struct wee_linux_kludge_pkthdr));
		}
		if((h->caplen > 65535) || (h->caplen < 0))
		{
			if(linux_kludge == 0)
				error("Bogus caplen; try using -k");
			error("Bogus caplen.  Ignominiously admitting defeat.");
		}
		if((bar = gzread(zfile, sp, h->caplen)) < 0)
			error("Bad packet read");
		if(bpf_filter(base_fcode.bf_insns, sp, h->len, h->caplen))
		{
			pcap_print_pt(NULL, hh, sp);
		}
	}
	gzclose(zfile);
	free(h);
	free(hh);
	free(sp);
	free(kp);
}

char *
get_filter(char **argv)
{
	char **p;
	int i = 0;
	char *foo, *bar, *baz;

	p = argv;
	if(*p == NULL) return(NULL);
	while(*p)
		i += strlen(*p++) + 1;
	if(!(foo = (char *)calloc(1, i)))
		error("calloc:  get_filter");
	bar = foo;
	p = argv;
	while((baz = *p++) != NULL)
	{
		while((*bar++ = *baz++) != '\0');
		bar[-1] = ' ';
	}
	bar[-1] = '\0';
	strncpy(blah, foo, MAX_FILTER_LEN);
	free(foo);
	return(blah);
}

/*
 This just creates a file containing a local pcap header.  The file
 is used by some of the things that do offline reading.
 The entire exercise is basically just a crude way to get around
 the semantics of the filter compiling code native to libpcap.  The
 Right Way To Do It is probably to rewrite the filter compilers, but
 I'm one lazy bastard and this was quicker.	-spb
*/ 
void
make_fake(char *conf)
{
	pcap_t *pd;
	pcap_dumper_t *foo;

	if(!(pd = (pcap_t *)calloc(1, sizeof(pcap_t))))
		error("calloc:  make_fake");
	pd->linktype = DLT_EN10MB;		/* Fuck FDDI */
	pd->tzoff = 0;
	pd->snapshot = default_snaplen;
	foo = pcap_dump_open(pd, conf);
	pcap_dump_close(foo);
	free(pd);
}

void
default_formats(int d_f[], char *d_d[], int d_t[])
{
	int i;
	i = 0;
	if(pf_outformat[0] == 0)
	{
		while(d_f[i] != 0)
		{
			pf_outformat[i] = d_f[i];
			i++;
		}
	}
	i = 0;
	if(pf_outdelim[0] == 0)
	{
		while(d_d[i] != NULL)
		{
			pf_outdelim[i] = d_d[i];
			i++;
		}
	}
	i = 0;
	while(d_t[i] != -1)
	{
		pf_type[i] = d_t[i];
		i++;
	}
}

